@@ -132,9 +132,16 @@ static int rsync_module(int fd, int i)
 	}
 
 	if (!claim_connection(lp_lock_file(), lp_max_connections())) {
-		rprintf(FERROR,"max connections (%d) reached\n",
-			lp_max_connections());
-		io_printf(fd,"@ERROR: max connections (%d) reached - try again later\n", lp_max_connections());
+		if (errno) {
+			rprintf(FERROR,"failed to open lock file %s : %s\n",
+				lp_lock_file(), strerror(errno));
+			io_printf(fd,"@ERROR: failed to open lock file %s : %s\n",
+				  lp_lock_file(), strerror(errno));
+		} else {
+			rprintf(FERROR,"max connections (%d) reached\n",
+				lp_max_connections());
+			io_printf(fd,"@ERROR: max connections (%d) reached - try again later\n", lp_max_connections());
+		}
 		return -1;
 	}
 
